@@ -1230,7 +1230,8 @@ int gsm411_send_sms(struct gsm_network *net,
 int gsm411_send_rp_data(struct gsm_network *net, struct vlr_subscr *vsub,
 			size_t sm_rp_oa_len, const uint8_t *sm_rp_oa,
 			size_t sm_rp_ud_len, const uint8_t *sm_rp_ud,
-			bool sm_rp_mmts_ind)
+			bool sm_rp_mmts_ind, const uint8_t *gsup_source_name,
+			size_t gsup_source_name_len)
 {
 	struct gsm_trans *trans;
 	struct msgb *msg;
@@ -1245,6 +1246,17 @@ int gsm411_send_rp_data(struct gsm_network *net, struct vlr_subscr *vsub,
 	if (trans->msc_a != NULL)
 		gsm411_handle_mmts_ind(trans);
 
+	/* Save GSUP source_name for subsequent response messages */
+	if (gsup_source_name && gsup_source_name_len) {
+		trans->sms.gsup_source_name = talloc_memdup(trans, gsup_source_name,
+							    gsup_source_name_len);
+		if (!trans->sms.gsup_source_name) {
+			trans_free(trans);
+			return -ENOMEM;
+		}
+		trans->sms.gsup_source_name_len = gsup_source_name_len;
+	}
+
 	/* Allocate a message buffer for to be encoded SMS */
 	msg = gsm411_msgb_alloc();
 	if (!msg) {
