@@ -1,5 +1,5 @@
 /* MSC interface to quagga VTY */
-/* (C) 2016-2018 by sysmocom s.m.f.c. GmbH <info@sysmocom.de>
+/* (C) 2016-2018 by sysmocom s.f.m.c. GmbH <info@sysmocom.de>
  * Based on OpenBSC interface to quagga VTY (libmsc/vty_interface_layer3.c)
  * (C) 2009-2017 by Harald Welte <laforge@gnumonks.org>
  * (C) 2009-2011 by Holger Hans Peter Freyther
@@ -68,6 +68,7 @@
 #include <osmocom/msc/sccp_ran.h>
 #include <osmocom/msc/ran_peer.h>
 #include <osmocom/msc/ran_infra.h>
+#include <osmocom/msc/asci_vty.h>
 
 static struct gsm_network *gsmnet = NULL;
 
@@ -131,19 +132,23 @@ DEFUN(cfg_net_mnc,
 
 DEFUN(cfg_net_name_short,
       cfg_net_name_short_cmd,
-      "short name NAME",
+      "short name .NAME",
       "Set the short GSM network name\n" NAME_CMD_STR NAME_STR)
 {
-	osmo_talloc_replace_string(gsmnet, &gsmnet->name_short, argv[0]);
+	if (gsmnet->name_short != NULL)
+		talloc_free(gsmnet->name_short);
+	gsmnet->name_short = argv_concat(argv, argc, 0);
 	return CMD_SUCCESS;
 }
 
 DEFUN(cfg_net_name_long,
       cfg_net_name_long_cmd,
-      "long name NAME",
+      "long name .NAME",
       "Set the long GSM network name\n" NAME_CMD_STR NAME_STR)
 {
-	osmo_talloc_replace_string(gsmnet, &gsmnet->name_long, argv[0]);
+	if (gsmnet->name_long != NULL)
+		talloc_free(gsmnet->name_long);
+	gsmnet->name_long = argv_concat(argv, argc, 0);
 	return CMD_SUCCESS;
 }
 
@@ -2071,13 +2076,14 @@ void msc_vty_init(struct gsm_network *msc_network)
 	osmo_tdef_vty_groups_init(MSC_NODE, msc_tdef_group);
 
 	/* Deprecated: Old MGCP config without pooling support in MSC node: */
-	mgcp_client_vty_init(msc_network, MSC_NODE, &msc_network->mgw.conf);
+	mgcp_client_vty_init(msc_network, MSC_NODE, msc_network->mgw.conf);
 
 #ifdef BUILD_IU
 	ranap_iu_vty_init(MSC_NODE, (enum ranap_nsap_addr_enc*)&msc_network->iu.rab_assign_addr_enc);
 #endif
 	sgs_vty_init();
 	smsc_vty_init(msc_network);
+	asci_vty_init(msc_network);
 
 	osmo_fsm_vty_add_cmds();
 
