@@ -1,5 +1,5 @@
 /*
- * (C) 2019 by sysmocom - s.m.f.c. GmbH <info@sysmocom.de>
+ * (C) 2019 by sysmocom - s.f.m.c. GmbH <info@sysmocom.de>
  * All Rights Reserved
  *
  * SPDX-License-Identifier: AGPL-3.0+
@@ -33,6 +33,7 @@
 #include <osmocom/msc/vlr.h>
 #include <osmocom/msc/ran_conn.h>
 #include <osmocom/msc/cell_id_list.h>
+#include <osmocom/msc/msc_vgcs.h>
 
 static struct osmo_fsm ran_peer_fsm;
 
@@ -118,7 +119,12 @@ void ran_peer_discard_all_conns(struct ran_peer *rp)
 	struct ran_conn *conn, *next;
 
 	ran_peer_for_each_ran_conn_safe(conn, next, rp) {
-		ran_conn_discard(conn);
+		/* Tell VGCS FSM that the connections have been cleared. */
+		if (conn->vgcs.bss)
+			vgcs_vbs_clear_cpl(conn->vgcs.bss, NULL);
+		else if (conn->vgcs.cell)
+			vgcs_vbs_clear_cpl_channel(conn->vgcs.cell, NULL);
+		else ran_conn_discard(conn);
 	}
 }
 
@@ -393,18 +399,22 @@ void ran_peer_st_ready(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 		OSMO_ASSERT(ctx->conn);
 		OSMO_ASSERT(ctx->msg);
 
-		if (!ctx->conn->msc_role) {
+		if (ctx->conn->msc_role) {
+			/* "normal" A connection, dispatch to MSC-I or MSC-T */
+			an_apdu = (struct an_apdu){
+				.an_proto = rp->sri->ran->an_proto,
+				.msg = ctx->msg,
+			};
+			osmo_fsm_inst_dispatch(ctx->conn->msc_role, MSC_EV_FROM_RAN_UP_L2, &an_apdu);
+		} else if (ctx->conn->vgcs.bss) {
+			/* VGCS call related */
+			msc_a_rx_vgcs_bss(ctx->conn->vgcs.bss, ctx->conn, ctx->msg);
+		} else if (ctx->conn->vgcs.cell) {
+			/* VGCS channel related */
+			msc_a_rx_vgcs_cell(ctx->conn->vgcs.cell, ctx->conn, ctx->msg);
+		} else
 			LOG_RAN_PEER(rp, LOGL_ERROR,
 				     "Rx CO message on conn that is not associated with any MSC role\n");
-			return;
-		}
-
-		an_apdu = (struct an_apdu){
-			.an_proto = rp->sri->ran->an_proto,
-			.msg = ctx->msg,
-		};
-
-		osmo_fsm_inst_dispatch(ctx->conn->msc_role, MSC_EV_FROM_RAN_UP_L2, &an_apdu);
 		return;
 
 	case RAN_PEER_EV_MSG_DOWN_CO_INITIAL:
